import asyncio
import json
import pprint

from importlib import import_module

import bs4
import pytest

from channels.testing import HttpCommunicator
from django.contrib.auth import login
from django.http import HttpRequest
from django.core.management import call_command
from django.middleware import csrf
from rest_framework import status

from bdd_coder import decorators
from bdd_coder import tester

from django_tasks.task_runner import TaskRunner
from django_tasks.websocket_client import LocalWebSocketClient


@pytest.mark.django_db
class BddTester(tester.BddTester):
    """
    The BddTester subclass of this tester package.
    It manages scenario runs. All test classes inherit from this one,
    so generic test methods for this package are expected to be defined here
    """
    gherkin = decorators.Gherkin(logs_path='bdd_runs.log')
    runner = TaskRunner.get()

    task_durations = [0.995, 0.95, 0.94, 0.8]
    credentials = dict(username='Alice', password='AlicePassWd')

    @pytest.fixture(autouse=True)
    def setup_ws_client(self, event_loop):
        self.ws_client = LocalWebSocketClient(timeout=10)
        self.event_collection_task = self.ws_client.collect_events(event_loop)

    @pytest.fixture(autouse=True)
    def setup_asgi_models(self, settings):
        settings.ALLOWED_HOSTS = ['*']
        # settings.MIDDLEWARE.insert(3, 'django_tasks.behaviour.tests.DisableCSRFMiddleware')
        settings.SESSION_SAVE_EVERY_REQUEST = True
        settings.MIDDLEWARE.insert(1, 'django_tasks.behaviour.tests.AuthTestMiddleware')
        self.settings = settings
        from django_tasks import asgi, models

        self.api_asgi = asgi.http_paths[0].callback
        self.admin_asgi = asgi.http_paths[1].callback
        self.models = models

    def store_session_cookie(self, user):
        # Create a fake request to store login details.
        self.request = HttpRequest()
        self.session_engine = import_module(self.settings.SESSION_ENGINE)
        self.request.session = self.session_engine.SessionStore()
        login(self.request, user)

        # Save the session values.
        self.request.session.save()

        # Set the cookie to represent the session.
        self.request.COOKIES[self.settings.SESSION_COOKIE_NAME] = self.request.session.session_key

        # Set the CSRF cookie
        csrf._add_new_csrf_cookie(self.request)
        self.request.COOKIES['csrftoken'] = self.request.META['CSRF_COOKIE']

        from django.contrib.sessions.models import Session
        print(self.request.session.load(), Session.objects.all())

    async def assert_admin_call(self, method, path, expected_http_code, data=None):
        headers = [(b'CONTENT_TYPE', b'application/x-www-form-urlencoded')]

        if self.request.COOKIES:
            cookie_header = '; '.join(f'{k}={v}' for k, v in self.request.COOKIES.items())
            headers.append((b'COOKIE', cookie_header.encode()))

        data = data or {}

        if method.lower() not in ['get', 'head', 'options', 'trace']:
            data['csrfmiddlewaretoken'] = csrf.get_token(self.request)

        body = '&'.join([f'{k}={v}' for k, v in data.items()]).encode()

        responses = await self.assert_daphne_call(
            self.admin_asgi, method, path, expected_http_code, body, headers)

        return responses

    async def assert_rest_api_call(self, method, api_path, expected_http_code, json_data=None):
        body, headers = b'', [(b'HTTP_AUTHORIZATION', f'Token {self.get_output("token")}'.encode())]

        if json_data:
            headers.append((b'CONTENT_TYPE', b'application/json'))
            body = json.dumps(json_data).encode()

        responses = await self.assert_daphne_call(
            self.api_asgi, method, api_path, expected_http_code, body, headers
        )
        return responses

    @classmethod
    async def assert_daphne_call(cls, asgi, method, path, expected_http_code, body=b'', headers=None):
        communicator = HttpCommunicator(asgi, method, path, body=body, headers=headers)
        response = await communicator.get_response()

        if response['status'] == status.HTTP_302_FOUND:
            redirected_responses = await cls.assert_daphne_call(
                asgi, 'GET', cls.get_response_header(response, 'Location'), expected_http_code
            )
            return [response, *redirected_responses]

        assert response['status'] == expected_http_code
        return [response]

    @staticmethod
    def get_response_header(response, header_name):
        header, value = next(filter(lambda h: h[0].decode().lower() == header_name.lower(), response['headers']))
        return value.decode()

    async def fake_task_coro_ok(self, duration):
        await asyncio.sleep(duration)
        return duration

    async def fake_task_coro_raise(self, duration):
        await asyncio.sleep(duration)
        raise Exception('Fake error')

    def get_all_admin_messages(self, soup):
        return {k: self.get_admin_messages(soup, k) for k in ('success', 'warning', 'info')}

    @staticmethod
    def get_admin_messages(soup, message_class):
        return [li.contents[0] for li in soup.find_all('li', {'class': message_class})]

    @staticmethod
    def get_soup(content):
        return bs4.BeautifulSoup(content.decode(), features='html.parser')

    def a_tasks_admin_user_is_created_with_command(self, django_user_model):
        self.credentials['password'] = call_command(
            'create_task_admin', self.credentials['username'], 'fake@gmail.com'
        )
        user = django_user_model.objects.get(username=self.credentials['username'])
        self.store_session_cookie(user)

        return user,

    async def cancelled_error_success_messages_are_broadcasted(self):
        cancelled, error, success = map(int, self.param)
        self.ws_client.expected_events = {
            'started': cancelled + error + success,
            'cancelled': cancelled, 'error': error, 'success': success,
        }
        timeout = 2
        try:
            await asyncio.wait_for(self.event_collection_task, timeout)
        except TimeoutError:
            self.ws_client.wsapp.close()
            raise AssertionError(
                f'Timeout in event collection. Expected counts: {self.ws_client.expected_events}. '
                f'Collected events in {timeout}s: {pprint.pformat(self.ws_client.events)}.')
        else:
            self.ws_client.expected_events = {}



import asyncio
import time

from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncJsonWebsocketConsumer
from django.conf import settings
from django.core.cache import cache
from rest_framework import exceptions, status

from django_tasks.drf_consumer import DrfConsumer
from django_tasks.doctask_scheduler import DocTaskScheduler
from django_tasks.serializers import DocTaskSerializer, TaskEventSerializer
from django_tasks.task_inspector import get_coro_info
from django_tasks.task_runner import TaskRunner


class TasksRestConsumer(DrfConsumer, DocTaskScheduler):
    async def process_drf_response(self, drf_response):
        if drf_response.status_code == status.HTTP_201_CREATED:
            if isinstance(drf_response.data, dict):
                await self.schedule_doctask(drf_response.data)
            elif isinstance(drf_response.data, list):
                await asyncio.gather(*[self.schedule_doctask(data) for data in drf_response.data])


class TaskEventsConsumer(AsyncJsonWebsocketConsumer, DocTaskScheduler):
    groups = [settings.CHANNEL_TASKS.channel_group]

    async def task_started(self, event):
        """Handles the DocTask, if any, and echoes the task.started document."""
        await self.handle_doctask(event['content'])
        await self.distribute_task_event(event)

    async def task_success(self, event):
        """Echoes the task.success document."""
        await self.distribute_task_event(event)

    async def task_cancelled(self, event):
        """Echoes the task.cancelled document."""
        await self.distribute_task_event(event)

    async def task_error(self, event):
        """Echoes the task.error document."""
        await self.distribute_task_event(event)

    async def task_badrequest(self, event):
        """Echoes the task.badrequest document."""
        await self.distribute_task_event(event)

    async def handle_doctask(self, event_content):
        doctask = await database_sync_to_async(self.retrieve_doctask)(event_content['memory-id'])
        if doctask:
            doctask.document.append(event_content)
            await doctask.asave()

    async def distribute_task_event(self, event):
        event['timestamp'] = time.time()
        await self.send_json(content=event)
        self.cache_task_event(event)

    def cache_task_event(self, event):
        cache_key = f"{self.scope['user'].username}.task_events"
        user_task_events = cache.get(cache_key, {})
        user_task_events[str(event['timestamp'])] = event['content']
        cache.set(cache_key, user_task_events)

    def clear_task_cache(self, content):
        memory_id = content.get('memory-id', 0)
        cache_key = f"{self.scope['user'].username}.task_events"
        cache.set(cache_key, {
            timestamp: data for timestamp, data in cache.get(cache_key, {}).items()
            if data['memory-id'] != memory_id})

    async def receive_json(self, event):
        """Pocesses task schedule websocket requests, and task cache clear requests."""
        serializer = TaskEventSerializer(data=event)

        try:
            serializer.is_valid(raise_exception=True)
        except exceptions.ValidationError as error:
            await self.send_bad_request_message(error)
        else:
            if serializer.data['type'] == 'task.clear':
                self.clear_task_cache(serializer.data['content'])
            elif serializer.data['type'] == 'task.schedule':
                await self.schedule_tasks(serializer.data['content'])

    async def send_bad_request_message(self, error: exceptions.ValidationError):
        await self.channel_layer.group_send(settings.CHANNEL_TASKS.channel_group, {
            'type': 'task.badrequest', 'content': {
                'details': error.get_full_details(), 'status': 'BadRequest'
            }
        })

    async def schedule_tasks(self, content):
        """Pocesses task schedule websocket requests, and task cache clear requests."""
        try:
            many_serializer = await database_sync_to_async(DocTaskSerializer.get_task_group_serializer)(content)
        except exceptions.ValidationError as error:
            await self.send_bad_request_message(error)
        else:
            runner = TaskRunner.get()
            await asyncio.gather(*[runner.schedule(
                get_coro_info(task['registered_task'], **task['inputs']).callable(**task['inputs'])
            ) for task in many_serializer.data])



from importlib import import_module

import logging

import pytest

from django.conf import settings
from django.contrib.auth import login
from django.contrib.sessions.models import Session

pytest.register_assert_rewrite(f'{__name__}.base')


class DisableCSRFMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        setattr(request, '_dont_enforce_csrf_checks', True)
        response = self.get_response(request)
        return response


class AuthTestMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self.logger = logging.getLogger('django')
        self.session_engine = import_module(settings.SESSION_ENGINE)

    def __call__(self, request):
        response = self.get_response(request)

        self.log_object(request, 'COOKIES')
        self.log_object(request, 'META')
        self.log_object(request.session, 'session_key')
        request.session.save()
        self.log_object(request, 'session', lambda s: s.load())
        self.log_object(Session.objects, 'all')

        return response

    def log_object(self, obj, key='', f=lambda v: v):
        v = getattr(obj, key, None) if hasattr(obj, key) else (obj if not key else None)
        self.logger.debug('%s.%s: %s', type(obj).__name__, key, f(v() if callable(v) else v))



import configparser
import os


class SettingsIni:
    ini_key = 'CHANNEL_TASKS_INI_PATH'
    secret_key_key = 'DJANGO_SECRET_KEY'
    default_installed_apps = [
        'bootstrap5',
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.staticfiles',
        'rest_framework.authtoken',
        'rest_framework',
        'django.contrib.messages',
        'django_extensions',
        'django_filters',
        'django_tasks',
        'django.contrib.admin',
    ]

    def __init__(self):
        ini_path = os.getenv(self.ini_key, '')
        assert os.path.isfile(ini_path), f'Channel-tasks settings file at {self.ini_key}={ini_path} not found.'
        self.ini = configparser.ConfigParser()
        self.ini.read(ini_path)

        assert self.secret_key_key in os.environ, f'Expected a Django secret key in {self.secret_key_key} envvar.'
        self.secret_key = os.environ[self.secret_key_key]

    def get_array(self, section, key, default):
        return ([line.strip() for line in self.ini[section][key].splitlines() if line.strip()]
                if self.ini.has_option(section, key) else default)

    def get_boolean(self, section, key, default):
        return self.ini[section].getboolean(key, default) if self.ini.has_section(section) else default

    def get_int(self, section, key, default):
        return self.ini[section].getint(key, default) if self.ini.has_section(section) else default

    def get_text(self, section, key, default):
        return self.ini[section][key].strip() if self.ini.has_option(section, key) else default

    @property
    def allowed_hosts(self):
        return ['127.0.0.1', self.server_name]

    @property
    def install_apps(self):
        return self.get_array('apps', 'install-apps', [])

    @property
    def debug(self):
        return self.get_boolean('security', 'debug', False)

    @property
    def server_name(self):
        return self.get_text('security', 'server-name', 'localhost')

    @property
    def proxy_route(self):
        return self.get_text('security', 'proxy-route', '')

    @property
    def local_port(self):
        return self.get_int('security', 'local-port', 8001)

    @property
    def log_level(self):
        return self.get_text('logging', 'log-level', 'INFO')

    @property
    def expose_doctask_api(self):
        return self.get_boolean('asgi', 'expose-doctask-api', False)

    @property
    def databases(self):
        if not self.ini.has_section('database'):
            return {
                'default': {
                    'ENGINE': 'django.db.backends.sqlite3',
                    'NAME': 'channel-tasks.sqlite3',
                }
            }

        default = {k.upper(): v for k, v in dict(self.ini['database']).items()}
        default.setdefault('PASSWORD', os.getenv('CHANNEL_TASKS_DB_PASSWORD', ''))

        return {'default': default}

    @property
    def channel_layers(self):
        return {
            'default': {
                'BACKEND': 'channels_redis.core.RedisChannelLayer',
                'CONFIG': {
                    'hosts': [(self.redis_host, self.redis_port)],
                },
            },
        }

    @property
    def caches(self):
        return {
            'default': {
                'BACKEND': 'django.core.cache.backends.redis.RedisCache',
                'LOCATION': f'redis://{self.redis_host}:{self.redis_port}',
                'TIMEOUT': 4*86400,
            },
        }

    @property
    def redis_host(self):
        return self.get_text('redis', 'host', '127.0.0.1')

    @property
    def channel_group(self):
        return self.get_text('redis', 'channel-group', 'tasks')

    @property
    def redis_port(self):
        return self.get_int('redis', 'port', 6379)

    @property
    def static_root(self):
        return self.get_text('security', 'static-root', '/www/django_tasks/static')

    @property
    def media_root(self):
        return self.get_text('security', 'media-root', '/www/django_tasks/media')

    @property
    def email_settings(self):
        return (self.email_host,
                self.email_port,
                self.email_use_tls,
                os.getenv('CHANNEL_TASKS_EMAIL_USER', ''),
                os.getenv('CHANNEL_TASKS_EMAIL_PASSWORD', ''))

    @property
    def email_host(self):
        return self.get_text('email', 'host', '')

    @property
    def email_port(self):
        return self.get_int('email', 'port', 0)

    @property
    def email_use_tls(self):
        return self.get_boolean('email', 'use-tls', False)

    def sort_installed_apps(self, *apps: list[str]) -> list[str]:
        return self.default_installed_apps + [
            k for k in apps if k not in self.default_installed_apps] + self.install_apps
